/*
 * system time synchronization for VMware guest OS
 *     by <osam-a@astroarts.co.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/time.h>
#include <sys/stat.h>

char *myname = "vmtimesync";
#define VERSION "1.0"

typedef unsigned int uint32;

#define VM_MAGIC	0x564D5868
#define VM_PORT		0x5658

#define DEFAULT_INTERVAL	5
#define DEFAULT_DIFFERENCE	200

//
// options
//
int opt_debug = 0;
int opt_interval = DEFAULT_INTERVAL;
int opt_difference = DEFAULT_DIFFERENCE; // ms

//
// setjmp
//
int jumpok = 0;
sigjmp_buf jmpbuf;

//
// signal function replacement
//
void *sig_set_handler(int signo, void *func)
{
	struct sigaction act, oact;
	
	act.sa_handler = func;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	if (sigaction(signo, &act, &oact) < 0) {
		return SIG_ERR;
	}
	return oact.sa_handler;
}

//
// catch segmentation fault exception
//
void sig_catch_segfault()
{
	if (jumpok) {
		jumpok = 0;
		siglongjmp(jmpbuf, 1);
	}
}

//
// get VMware version
//
void vm_get_version(uint32 *pversion, uint32 *pmagic)
{
	sig_set_handler(SIGSEGV, sig_catch_segfault);
	if (sigsetjmp(jmpbuf, 1)) {
		*pversion = *pmagic = 0;
		return;
	}
	jumpok = 1;
	
	__asm__ __volatile__
		(
		 "pushl %%ebx\n\t"
		 "pushl %%ecx\n\t"
		 "pushl %%edx\n\t"
		 
		 "movl $10, %%ecx\n\t"
		 "inl %w3,%0\n\t"
		 "movl %%ebx, %1\n\t"
		 
		 "popl %%edx\n\t"
		 "popl %%ecx\n\t"
		 "popl %%ebx\n\t"
		 
		 : "=a" (*pversion),
		   "=m" (*pmagic)
		 : "a" (VM_MAGIC),
		   "d" (VM_PORT)
		 : "%ebx",
		   "%ecx"
		 );
	
	sig_set_handler(SIGSEGV, SIG_DFL);
}

//
// get system time of host OS
//
void vm_get_host_time(uint32 *psec, uint32 *pusec, uint32 *ptz)
{
	__asm__ __volatile__
		(
		 "pushl %%ebx\n\t"
		 "pushl %%ecx\n\t"
		 "pushl %%edx\n\t"
		 
		 "movl $0x17, %%ecx\n\t"
		 "inl %w4,%0\n\t"
		 "movl %%ebx, %1\n\t"
		 "movl %%edx, %2\n\t"
		 
		 "popl %%edx\n\t"
		 "popl %%ecx\n\t"
		 "popl %%ebx\n\t"
		 
		 : "=a" (*psec),
		   "=m" (*pusec),
		   "=m" (*ptz)
		 : "a" (VM_MAGIC),
		   "d" (VM_PORT)
		 : "%ebx",
		   "%ecx"
		 );
}

//
// show usage
//
void usage()
{
	fprintf(stderr, "vmtimesync version %s by <osam-a@astroarts.co.jp>\n",
			VERSION);
	fprintf(stderr, "Usage: %s [-n] [-i <interval>] [-d <max difference>] [-h]\n",
			myname);
	fprintf(stderr, "  <interval> should be more than 0, default is %d(s).\n",
			DEFAULT_INTERVAL);
	fprintf(stderr, "  <max difference> should be more than 10, default is %d(ms).\n",
			DEFAULT_DIFFERENCE);
}

//
// main
//
int main(int argc, char **argv)
{
	uint32 version, magic;
	pid_t pid;
	char c;
	
	while ((c = getopt(argc, argv, "ni:d:h")) != EOF) {
		switch (c) {
		case 'n':
			opt_debug = 1;
			break;
		case 'i':
			if ((opt_interval = atoi(optarg)) < 1) {
				usage();
				return 1;
			}
			break;
		case 'd':
			if ((opt_difference = atoi(optarg)) < 10) {
				usage();
				return 1;
			}
			break;
		case 'h':
		default:
			usage();
			return 1;
		}
	}
	
	if (geteuid() != 0 && ! opt_debug) {
		fprintf(stderr, "%s: you must be root to run\n", myname);
		return 1;
	}
	
	// sure to run inside VMware
	vm_get_version(&version, &magic);
	if (magic != VM_MAGIC) {
		fprintf(stderr, "%s: not running inside VMware\n", myname);
		return 1;
	}
	
	// fork process and get into daemon
	if (! opt_debug) {
		if ((pid = fork()) < 0) {
			fprintf(stderr, "%s: fork failed\n", myname);
			return 1;
		} else if (pid != 0) {
			return 0;
		}
		setsid();
		chdir("/");
		umask(0);
	}
	
	while (1) {
		uint32 sec, usec, tzmin;
		struct timeval tv;
		struct timezone tz;
		int diff;
		
		// time of host OS and virtual machine
		vm_get_host_time(&sec, &usec, &tzmin);
		gettimeofday(&tv, &tz);
		
		// get time differnce in ms (abs)
		if (sec > tv.tv_sec) {
			diff = ((int)(sec - tv.tv_sec) * 1000)
				+ ((int)usec - (int)tv.tv_usec) / 1000;
		} else if (sec < tv.tv_sec) {
			diff = ((int)(tv.tv_sec - sec) * 1000)
				+ ((int)tv.tv_usec - (int)usec) / 1000;
		} else {
			diff = abs(((int)tv.tv_usec - (int)usec) / 1000);
		}
		
		// debug messages
		if (opt_debug) {
			printf("VM: %d.%06d (TZ:%d)\n",
				   sec, usec, tzmin);
			printf("OS: %d.%06d (TZ:%d)\n",
				   (int)tv.tv_sec, (int)tv.tv_usec, tz.tz_minuteswest);
			printf("difference: %d [ms]\n", diff);
		}
		
		// sync if needed
		if (diff >= opt_difference) {
			tv.tv_sec  = sec;
			tv.tv_usec = usec;
			if (opt_debug) {
				printf("SYNC!\n");
			} else {
				settimeofday(&tv, &tz);
			}
		}
		sleep(opt_interval);
	}
	
	return 0;
}
